open Unix;;

(*
let () = Unix.handle_unix_error (fun () -> print_endline (Unix.getcwd())) ()
*)

let equal_node n1 n2 =
  n1.st_ino = n2.st_ino && n1.st_dev = n2.st_dev

let try_finalize f x finally y =
  let res = try f x with exn -> finally y; raise exn in
  finally y;
  res

let dir_find f path =
  let dir_handle = opendir path in
  let rec find () =
    let name =
      try readdir dir_handle
      with End_of_file -> raise Not_found
    in
    if f name then name else find ()
  in
  try_finalize
    find ()
    closedir dir_handle

(* L'appel système Unix.lstat est coûteux, on évite donc de répéter
   plusieurs fois. Pour cela, on va maintenir pour chaque noeud une
   structure qui regroupe le chemin et l'information sur le noeud associé
   au chemin.
*)

type info = { path : string; lstat : stats }

let info path = { path = path; lstat = lstat path }

(* Voici la fonction mon_getcwd, elle peut lever l'exception Not_found
   lorsqu'on ne réussit pas à trouver de chemin, ce qui peut se produire
   si l'arborescence a été modifiée en cours de calcul (par un autre
   programme s'exécutant en parallèle).

   Noter qu'il est inutile de tester current_dir_name et
   parent_dir_name car on sait à coup sûr qu'ils ne peuvent pas
   convenir. On évite ainsi deux appels systèmes inutiles.
*)

let mon_getcwd =
  let rec find_root node =
    let parent_node =
      info (Filename.concat Filename.parent_dir_name node.path)
    in
    if equal_node node.lstat parent_node.lstat then "/"
    else
      let equal name =
        name <> Filename.current_dir_name &&
          name <> Filename.parent_dir_name &&
          equal_node
            node.lstat
            (Unix.lstat (Filename.concat parent_node.path name))
      in
      let name = dir_find equal parent_node.path in
      Filename.concat (find_root parent_node) name
  in
  fun () -> find_root (info Filename.current_dir_name)


let () = Unix.handle_unix_error (fun () -> print_endline (mon_getcwd())) ()
